/*
 *  (C) Copyright ECWMF http://www.ecmwf.int
 *
 *  This software is licensed under the terms of the Apache Licence Version 2.0
 *  which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 *  In applying this licence, ECMWF does not waive the privileges and immunities
 *  granted to it by virtue of its status as an intergovernmental organisation nor
 *  does it submit to any jurisdiction.
 *
 *  -------------------------------------------------
 *
 *  Catalogue grid
 *
 *  The catalogue grid is a generic widget that shows lists of catalogue items.
 *  It is designed to show a very large number of items in a performant manner.
 *  The list can have filters applied, eg text search, facets, categories.
 *
 *  Sizing:
 *  The grid is designed primarily to fit into a box of externally determined size.
 *  The size must be set using css.
 *  Using size_fits_content mode, the height will be internally adjusted as the
 *  number of displayed items changes (max-height should be set using css).
 */
define(['config', 'jquery', 'bus', 'mixins', './filter', 'ajax_queue', './products', /*'jquery_ui/button',*/ './grid.css', 'utils', 'container', 'input-filter', './filter_status', 'perfect-scrollbar'], function(ECMWF, $, Bus, mixins, Filter, Ajaxqueue){

var view_modes = ['groups', 'thumbs', 'details', 'list'];

$.widget('ui.grid', $.extend({

	options: {
		products: undefined,
		view_mode: null,         // set this to override localstorage
		text_filter_parent: null,
		container: true,         // the container plugin is used to size the main content area to fill the available space
		size_fits_content: false,// forces container=false
		on_click: $.noop,        // user clicked on a product. (To track selection changes, subscribe to list.selection)
		on_dblclick: $.noop,
		on_model_change: $.noop, // args of this callback subject to review. is called early: users should not make too many assumptions about view status.
		on_filter_change: $.noop,
		on_view_change: $.noop,
		on_view_ready: $.noop,
		enable_status: true,
		bus: Bus()
	},

	_create: function() {
		this.element.addClass('catalogue-grid');
	},

	_init: function() {
		var self = this, o = self.options, list = o.products;
		this.widgetName = 'grid';
		if(o.size_fits_content) o.container = false;

		self.n_filter_runs = 0; // TODO is only used by the page views, they should track this themselves and it can be removed from the Grid

		$.map(list.filters, this.bind(self.add_filter));

		/* '_pending' lists outstanding update types that need to be performed:
		 *
		 *    -sizing:       calculate number of visible rows and columns.
		 *    -filtering:    calculate number of visible items and update the 'filtered' array.
		 *    -view:
		 *        -dom creation: update the number of dom elements.
		 *        -assignment:   assign model items to dom elements.
		 *    -selection:    show selection in the view.
		 */
		self._pending = {sizing: true, view: true};
		$.extend(self._pending, {selection: list.selection.always || undefined, filtering: (!list.filtered) || undefined});

		self._storage = this.storage('grid');

		var head = P(self.element.ensure('.head'), this);
		self.outer = self.element.ensure('.outer').css({width: '100%', position: 'relative'});
		var foot = self.element.ensure('.foot');

		head.addMenuItems($.map(view_modes, function(i){ return {name: i, select: function(i){ self._set_view_mode(i.name); }}; }));

		//head.menu();
		head.buttons({on_change: o.on_view_change});

		self.widgets = {
			form_select   : (false && head.formSelect({view_modes: view_modes, view_mode: o.view_mode, on_select: function(a){ self._set_view_mode(a); }})),
			filter_input  : ((o.text_filter_parent && P(o.text_filter_parent)) || head).filter_input(list.lookup_filter_by_name('text')),
			foot          : foot,
			status        : foot.ensure('.status')
		}

		head.elem.br('.clear');

		self.container = self.outer.ensure('.scrollable').css({height: '100%'})
			// TODO delay scrollbar? possible performance issue - forces renderering?
			.perfectScrollbar({suppressScrollX: true, wheelSpeed: ECMWF.wheelspeed || 10})
			.scroll($.throttle(200, function(e){
				self.container.catalogueView && self.container.catalogueView('scroll', -self.container.parent().scrollTop());
			}))
			.ensure('.grid-items');

		self._ready = $.Deferred(function(df){
			setTimeout(df.resolve, 400);
		});

		self._resize = $.debounce(200, function() {
			var self = this;
			this.container.catalogueView && this.container.catalogueView('resize');
			delete this._pending.sizing;
			this._update();
			self.container.scroll();
		});

		set_mode();

		list.subscribe(self, self._on_filter = function() {
			// the filter was run

			self._pending.view = list.changed;
			self.widgets.status.text(list.filtered.length + ' matching items');

			delete self._pending.filtering;
			self._update();

			o.on_filter_change.apply(self);

			self.n_filter_runs++;
		});

		list.selection.subscribe(self, function() {
			self._update({selection: true});
		});

		list.on.items.subscribe(self, function() {
			var visible = this.filter_pass;
			list.filter(); // maybe should be done inside the list?
			self._update({filtering: true});
		});

		self.filter_status = o.enable_status
			? foot.find('.filter-status').orelse(function(){ return foot.div('.status'); })
				.filterStatus({
					filters: list.filters,
					on_disable_change: function(filter_name, disabled){
						console.log('on_disable_change: ' + filter_name);
						var f = list.lookup_filter_by_name(filter_name);
						f.disabled = disabled;
						self._update({filtering: true});
					},
					on_remove: function(filter_name){
						var filter = list.lookup_filter_by_name(filter_name);
						console.log('grid.on_remove: ' + filter_name, filter);

						filter.deactivate();
						switch(filter.name){
							case 'facets':
								self._update({filtering: true});
								break;
							case 'text':
								self.widgets.filter_input.find('input').val('');
								break;
							default:
								break;
						}
					}
				})
			: $([]);

		if(list.filtered) self._on_filter();

		o.bus
			.listen('product-thumbnail-open', function(_, model){
				window.open(ECMWF.options.ajax.backend.interactive_product.replace('PRODUCT', model.name).replace('PROJECTION', 'global'), '_blank');
			})
			.listen('filters-invalidated', function(){
				self._update({filtering: true});
			});

		o.container && self.element.container({
			layout: function(elem, options) {
				var expandable = options.expandable;

				return {
					layout: function vertical(size) {
						if(__DEV__ && !self.element.is(':visible')) console.warn('cannot layout hidden element');
						var height = self.element.height();
						if(!height) return;

						elem.children(':visible').not(expandable).each(function(){
							var child = $(this);
							if(child.css('float') !== 'left' && !this.container_exclude){
								height -= child.outerHeight(true);
							}
						});

						var s = $.extend(size, {height: height});
						expandable.css('height', s.height).attr({sizor: 'container.custom'}).resize().content().triggerHandler('layout', s);
					}
				};
			},
			detached: true,
			expandable: self.outer
		})
		self.element.on('layout', function(){
			self._update({sizing: true});
		});

		function set_mode(){
			self.container.catalogueView && self.container.catalogueView('freeze');
			__DEV__ && Bus().send('event-start', ['start get item list']);
			list.have.items.then(function(items) {
				__DEV__ && Bus().send('event-start', ['end get item list']);
				list.items = items;

				self._set_view_mode(o.view_mode || /*self._storage.get().view || */view_modes[0]);

				o.on_model_change(items);
				self._update();
			});
		}
	},

	_destroy: function() {
		var products = this.options.products;

		products.unsubscribe(this);
		products.selection.unsubscribe(this);
	},

	add_filter: function(f) {
		var self = this;

		// TODO the list should bind to the filter change events, not the grid
		if($.isFunction(f.fn)){
			f.events && f.events.bind('changed', function(){
				self._update({filtering: true});
			});
		}
	},

	filter: function() {
		this._update({filtering: true});
	},

	/*
	 * These properties need to be maintained:
	 *     - filtered            in grid._filter_list
	 *     - n_columns, n_rows   in view.resize
	 *     - cache.n_cells       in view.resize
	 *     - assigned            in view.post_resize (allocates array items) + in view._assign_fresh (after filter/scroll)
	 *     - visible_items       in view._assign_fresh
	 */
	_update: function(change) {
		var self = this, o = this.options;

		if(change) self._pending = $.extend({}, change, self._pending); // note values from property take precedence.

		$.each(['filtering', 'sizing', 'view', 'selection'], function(i, v) {
			if(self._pending[v]){
				switch(v) {
					case 'filtering':
						self._pending = $.extend({view: true}, self._pending);
						self._ready.done(self.bind(o.products.filter)); // this is asynchronous and we get no callback. TODO for consistency, all update stages should perhaps be async.

						return false; // stop updates for now and wait for _update to be called again.

					case 'sizing':
						self._pending = $.extend({view: true}, self._pending);
						self._resize();
						return false;

					case 'view':
						if((self._pending.view.to_hide && self._pending.view.to_hide.length) || (self._pending.view.to_show && self._pending.view.to_show.length)){
							self.container.catalogueView &&
								self.container.catalogueView('hide_show_items', {hide: self._pending.view.to_hide, show: self._pending.view.to_show});
						}
						delete self._pending.view;
						break;

					case 'selection':
						self.container.catalogueView && self.container.catalogueView('select_by_idx', o.products.selection.value);
						delete self._pending.selection;
						break;

					default:
						break;
				}
			}
		});
	},

	_set_view_mode: function(mode) {
		var self = this, o = this.options;

		if(false) self._storage.set($.extend(self._storage.get(), {view: mode}));

		self.container.catalogueView && self.container.catalogueView('destroy');

		switch(mode){
			case 'thumbs':
				require(['grid/views/thumbs'], function(){
					make_view();
				});
				break;
			case 'details':
				require(['grid/views/detail'], function(){
					make_view();
				});
				break;
			case 'list':
				require(['grid/views/list'], function(){
					make_view();
				});
				break;
			case 'groups':
				//require(['grid/views/groups2'], function(){
				require(['grid/views/groups-v'], function(){
					make_view();
				});
				break;
			default: return;
		}

		function make_view(){
			if(!self.container[mode]) return console.warn('no view for mode=', mode);

			self.container.catalogueView = self.container[mode];
			self.container.catalogueView({
				grid: self,
				products: o.products,
				click: function(a){
					console.log('grid.click');
					return o.on_click(a);
				},
				dblclick: function(a){
					return o.on_dblclick(a);
				},
				select: function(a){
					console.log('grid.select');
					o.products.selection.set(o.products.filtered[a.filtered_idx]);
				}
			});
		}
	},

	resize: function() {
		this._update({sizing: true});
		this.element.trigger('layout');
	},

	select: function(product) {
		console.warn('grid.select (removed)');
		console.trace();
	},

	next: function() {
		var list = this.options.products;

		var i = $.inArray(+list.selection.value, list.filtered);
		return (i > -1 && i < list.items.length - 1)
			? this._select(list.filtered[i + 1])
			: null;
	},

	prev: function() {
		var list = this.options.products;
		var i = $.inArray(+list.selection.value, list.filtered);
		return (i > -1 && i > 0)
			? this._select(list.filtered[i - 1])
			: null;
	},

	_lookup_idx_by_name: function(product) {
		var list = this.options.products;
		for(var i in list.items){
			if(list.items[i].name() === product) return i;
		}
	}
}, mixins));


/* not currently used
$.widget('ui.imageList', $.ui.filterableImageList, {

	options: {
		viewType: 'details'
	},

	_init: function() {
		var self = this;

		self._add_filter(function(item){
			return (!self.regex || item.title.match(self.regex) || item.keywords.match(self.regex));
		});

		self._add_filter_input(
			this.element
				.div('.thumbnail-list-filter').text('Filter: ')
					.input({type: 'text'})
		);

		this.element.a({href: 'xxx'}).text('details');
		this.element.span(' | ');
		this.element.a({href: 'xxx'}).text('thumbnails');

		this.element.append(this.container);

		self.get_thumbnails();
	}
});
*/

var P = function(elem, grid) {
	return new GridPrivate(elem, grid);
};

var GridPrivate = function(elem, grid) {
	this.elem = this.$ = $(elem);
	this.grid = grid;
};

GridPrivate.prototype = {
	elem: null,
	grid: null,
	menu_items: [],
	menu: function(o) {
		var self = this, menu;

		function make_menu(){
			return $.map(self.menu_items, function(i){ return '<li data-value="hello"><a href="#">' + i.name + '</a></li>'; }).join('');
		}

		return $('<div class="burger"><span class="layer"></span><span class="layer"></span><span class="layer"></span></div>').appendTo(this.elem).click(function(){
			console.log('menu.click');
			menu && menu.toggle();
			menu = menu || $('<ul>' + make_menu() + '</ul>')
				.appendTo($(this))
				.menu({
					select: function(event, ui){
						var item = self.menu_items[ui.item.index()];
						console.log('menu select', item);
						item.select && item.select(item);
					}
				});
		});
	},
	buttons: function(o) {
		var self = this;
		$.each(this.menu_items, function(i, item) {
			self.elem.div('.view-type-button')
				.button({label: ECMWF.ucfirst(item.name), icons: {primary: 'icon-' + item.name.toLowerCase()}, text: false})
				.click(function() {
					console.log('view_type.click', item);
					item.select && item.select(item);
					o.on_change && o.on_change();
				});
		});
	},
	addMenuItems: function(items) {
		if(Array.isArray(items))
			$.merge(this.menu_items, items)
		else
			this.menu_items.push(items);
	},
	formSelect: function(o) {
		this.elem.formSelect({options: o.view_modes})
			.val(o.view_mode)
			.change(function(_){
				var selected = $(this).find('option:selected');
				console.log(selected.text());
				o.view_mode = selected.text();
				o.on_select(o.view_mode);
			});
	},
	filter_input: function(filter) {
		var input, self = this;
		return input = this.elem
			.find('.thumbnail-list-filter').orelse(function(){ return self.elem.div(); })
				.inputFilter({label: '', on_filter: function(text){
					filter.enable().on_change(null, text);
				}});
	}
}

});

